import sys
try:
    from sumatrapdf import SumatraPDF
except:
    pass

def lowercase(str):
    return str.lower()

def simpleload(source):
    "load file, strip whitespace and \\n's" 
    try:
        f = open(source, 'r', encoding='utf-8')
        lines = f.readlines()
        f.close()
        indexlist = [l.strip() for l in lines]
    except:
        indexlist = []
    return indexlist

def findentrypos(entry = '', indexlist = [], listtypefirst = False):
    if len(indexlist) == 0:
        return 0
    else:
        if listtypefirst:   # first entries listed
            if indexlist[0] >= entry:
                return 0
            else:
                for i, e in enumerate(indexlist):
                    if e >= entry:
                        return i - 1
                return len(indexlist) - 1
        else:               # last entries listed
            page = 0
            not_found = True
            for last in indexlist:
                if last < entry:
                    page += 1
                else:
                    not_found = False
                    break
            if not_found:
                page -= 1
            return page

class Dic():
    'generic dic class'
    def lookup(self, entry):
        pass
    def close(self):
        pass

class SDic(Dic):
    """SumatraPDF dictionary"""
    def __init__(self, filename='', index=None):
        Dic.__init__(self)
        self.index = index
        self.sumatra = SumatraPDF(filename)

    def lookup(self, entry):
        self.sumatra.show(page = self.index.page(entry))

    def close(self):
        self.sumatra.close()

class Index():
    'generic index class'
    def __init__(self, source = '',
        loadmethod = simpleload,
        encifer = lambda x: x,
        indexlist = [],
        listtypefirst = False, 
        transf = lowercase):
        # indexlist is the index entry list
        # listtypefirst = False if the last entry on the page is indicated
        # listtypefirst = True if the first entry on the page is indicated
        # transf is the procedure that makes the input suit the list format
        # by default converts to lowercase
        # lowercase function is defined in this module
        # if a source file is indicated
        # its contents overrides all the rest
        # loadmethod : function that is used in load_from_file
        # by default accepts a string (file name)
        # and a function (method of encifering)
        # and returns a list (entry list)
        # simpleload(source) is defined in this module
        # accepts a file and returns a list
        # it makes the list of file lines stripping whitespace
        # loadmethod can be redefined
        # in order to handle more sophisticated file formats
        # encifer : the raw entry list may need `encifering'
        # such as removing diacritics, etc.
        # no encifering by default
        self.__indexlist = indexlist
        self.__listtypefirst = listtypefirst
        self.transf = transf
        self.loadmethod = loadmethod
        self.encifer = encifer
        if source:
            self.load_from_file(source)

    # auxiliary methods

    def getindexlist(self):
        'get entry list'
        return self.__indexlist

    def setindexlist(self, list):
        'set entry list'
        self.__indexlist = list

    def getlisttype(self):
        'return listtypefirst'
        return self.__listtypefirst
        
    def setlisttype(self, listtype):
        'set listtypefirst'
        self.__listtypefirst = listtype

    def load_from_file(self, source):
        'load using loadmethod'
        self.setindexlist([self.encifer(e)
            for e in self.loadmethod(source)])

    # The only user-end method
    def page(self, entry):
        'the page defined for *entry* by the indexlist'
        return findentrypos(self.transf(entry),
                self.getindexlist(), self.getlisttype())

class SimpleIndex(Index):
    'Shifted first page'

    def __init__(self, source = '',
            loadmethod = simpleload,
            encifer = lambda x: x,
            indexlist = [],
            listtypefirst = False, 
            transf=lowercase, 
            firstpage = 1):
        # firstpage : the page number corresponding to the entry #0
        # The rest is same as in Index.__init__()
        Index.__init__(self,
            source = source,
            loadmethod = loadmethod,
            encifer = encifer,
            indexlist = indexlist,
            listtypefirst = listtypefirst,
            transf = transf)
        self.__firstpage = firstpage

    def getfirstpage(self):
        'get first page'
        return self.__firstpage

    def setfirstpage(self, firstpage):
        'set first page'
        self.__firstpage = firstpage

    def page(self, entry):
        'entry #0 is on page {firstpage}'
        return Index.page(self, entry) + self.getfirstpage()

### Multiple volumes ###

class VDic(Dic):
    'multiple volumes'
    def __init__(self, vol):
        # vol : list of volumes of class Dic
        # they must have nonempty indices with the same wordlisttype
        Dic.__init__(self)
        self.vol = vol
        self.vindex = VIndex([v.index for v in vol])

    def lookup(self, entry):
        vol = self.vindex.vol(entry)
        self.vol[vol].lookup(entry)

    def close(self):
        'close files'
        for v in self.vol:
            v.close()

class VIndex():
    'index for multiple volumes'
    # is created from a list of nonempty indices
    # the main property is the volpage(word) function

    def __init__(self, vols, listtypefirst = False, transf = lowercase):
        # vols is the list of volume indices
        # define : self.volindex is the index of volumes as if they were pages
        # volindex.wordlist is the list of last/first entries in the volumes
        # vols are constituent indices
        # listtypefirst and trans are assigned to each subordinate index

        self.listtypefirst = listtypefirst
        self.transf = transf
        self.vols = vols
        for v in self.vols:
            v.setlisttype(listtypefirst)
            v.transf = self.transf
        if listtypefirst:
            volindex = [i.getindexlist()[0] for i in self.vols]
        else:
            volindex = [i.getindexlist()[-1] for i in self.vols]
        self.volindex = Index(indexlist = volindex,
                listtypefirst = listtypefirst)

    def vol(self, entry):
        return self.volindex.page(entry) # `page' is volume number

    def volpage(self, entry):
        vol = self.volindex.page(entry)
        page = vol.page(entry)
        return (vol, page)
